@@ -370,25 +370,36 @@ gail_label_real_notify_gtk (GObject           *obj,
     }
   else if (strcmp (pspec->name, "cursor-position") == 0)
     {
-      gint start, end;
+      gint start, end, tmp;
       gboolean text_caret_moved = FALSE;
       gboolean selection_changed = FALSE;
 
       gail_obj = G_OBJECT (atk_obj);
       label = GTK_LABEL (widget);
 
+      if (gail_label->selection_bound != -1 && gail_label->selection_bound < gail_label->cursor_position)
+        {
+          tmp = gail_label->selection_bound;
+          gail_label->selection_bound = gail_label->cursor_position;
+          gail_label->cursor_position = tmp;
+        }
+
       if (gtk_label_get_selection_bounds (label, &start, &end))
         {
           if (start != gail_label->cursor_position ||
               end != gail_label->selection_bound)
             {
-              gail_label->selection_bound = end;
-              if (start != gail_label->cursor_position)
+              if (end != gail_label->selection_bound)
                 {
-                  /* cursor movement */
-                  gail_label->cursor_position = start;
-                  text_caret_moved = TRUE;
+                  gail_label->selection_bound = start;
+                  gail_label->cursor_position = end;
                 }
+              else
+                {
+                  gail_label->selection_bound = end;
+                  gail_label->cursor_position = start;
+                }
+              text_caret_moved = TRUE;
               if (start != end)
                 selection_changed = TRUE;
             }
@@ -399,12 +410,19 @@ gail_label_real_notify_gtk (GObject           *obj,
             selection_changed = TRUE;
           if (gtk_label_get_selectable (label))
             {
-              gint old_cursor_position = gail_label->cursor_position;
-
-              gail_label->cursor_position = start;
-              gail_label->selection_bound = end;
-              if (old_cursor_position != -1 && start != old_cursor_position)
+              if (gail_label->cursor_position != -1 && start != gail_label->cursor_position)
                 text_caret_moved = TRUE;
+              if (gail_label->selection_bound != -1 && end != gail_label->selection_bound)
+                {
+                  text_caret_moved = TRUE;
+                  gail_label->cursor_position = end;
+                  gail_label->selection_bound = start;
+                }
+              else
+                {
+                  gail_label->cursor_position = start;
+                  gail_label->selection_bound = end;
+                }
             }
           else
             {
@@ -414,7 +432,7 @@ gail_label_real_notify_gtk (GObject           *obj,
               gail_label->selection_bound = 0;
               text_caret_moved = TRUE;
             }
-            
+
         }
         if (text_caret_moved)
           g_signal_emit_by_name (gail_obj, "text_caret_moved", 
