@@ -20,13 +20,16 @@
 #include <gtk/gtk.h>
 #include "gailtextcell.h"
 #include "gailcontainercell.h"
-#include "gailtexthelper.h"
 #include "gailcellparent.h"
+#include <libgail-util/gailmisc.h>
+#include "gail-private-macros.h"
 
 static void      gail_text_cell_class_init		(GailTextCellClass *klass);
 static void      gail_text_cell_init			(GailTextCell	*text_cell);
 static void      gail_text_cell_finalize		(GObject	*object);
 
+static G_CONST_RETURN gchar* gail_text_cell_get_name    (AtkObject      *atk_obj);
+
 static void      atk_text_interface_init		(AtkTextIface	*iface);
 
 /* atktext.h */
@@ -54,22 +57,30 @@ static gchar*	gail_text_cell_get_text_after_offset	(AtkText	*text,
 static gint      gail_text_cell_get_character_count	(AtkText	*text);
 static gint      gail_text_cell_get_caret_offset	(AtkText	*text);
 static gboolean  gail_text_cell_set_caret_offset	(AtkText	*text,
-							gint		offset);
+							 gint		offset);
 static void      gail_text_cell_get_character_extents	(AtkText	*text,
-							gint		offset,
-							gint		*x,
-							gint		*y,
-							gint		*width,
-							gint		*height,
-							AtkCoordType	coords);
+							 gint		offset,
+							 gint		*x,
+							 gint		*y,
+							 gint		*width,
+							 gint		*height,
+							 AtkCoordType	coords);
 static gint      gail_text_cell_get_offset_at_point	(AtkText	*text,
-							gint		x,
-							gint		y,
-							AtkCoordType	coords);
-static AtkAttributeSet* gail_text_cell_ref_run_attributes (AtkText	*text,
-							gint		offset,
-							gint		*start_offset,      
-							gint		*end_offset); 
+							 gint		x,
+							 gint		y,
+							 AtkCoordType	coords);
+static AtkAttributeSet* gail_text_cell_get_run_attributes 
+                                                        (AtkText	*text,
+							 gint		offset,
+							 gint		*start_offset,      
+							 gint		*end_offset); 
+static AtkAttributeSet* gail_text_cell_get_default_attributes 
+                                                        (AtkText        *text);
+
+static PangoLayout*     create_pango_layout             (GtkCellRendererText *gtk_renderer,
+                                                         GtkWidget           *widget);
+static void             add_attr                        (PangoAttrList  *attr_list,
+                                                         PangoAttribute *attr);
 
 /* Misc */
 
@@ -122,34 +133,34 @@ gail_text_cell_get_type (void)
   static GType type = 0;
 
   if (!type)
-  {
-    static const GTypeInfo tinfo =
     {
-      sizeof (GailTextCellClass),
-      (GBaseInitFunc) NULL, /* base init */
-      (GBaseFinalizeFunc) NULL, /* base finalize */
-      (GClassInitFunc) gail_text_cell_class_init, /* class init */
-      (GClassFinalizeFunc) NULL, /* class finalize */
-      NULL, /* class data */
-      sizeof (GailTextCell), /* instance size */
-      0, /* nb preallocs */
-      (GInstanceInitFunc) gail_text_cell_init, /* instance init */
-      NULL /* value table */
-    };
+      static const GTypeInfo tinfo =
+      {
+        sizeof (GailTextCellClass),
+        (GBaseInitFunc) NULL, /* base init */
+        (GBaseFinalizeFunc) NULL, /* base finalize */
+        (GClassInitFunc) gail_text_cell_class_init, /* class init */
+        (GClassFinalizeFunc) NULL, /* class finalize */
+        NULL, /* class data */
+        sizeof (GailTextCell), /* instance size */
+        0, /* nb preallocs */
+        (GInstanceInitFunc) gail_text_cell_init, /* instance init */
+        NULL /* value table */
+      };
 
-    static const GInterfaceInfo atk_text_info =
-    {
+      static const GInterfaceInfo atk_text_info =
+      {
         (GInterfaceInitFunc) atk_text_interface_init,
         (GInterfaceFinalizeFunc) NULL,
         NULL
-    };
+      };
 
-    type = g_type_register_static (GAIL_TYPE_RENDERER_CELL,
-                                   "GailTextCell", &tinfo, 0);
-    g_type_add_interface_static (type, ATK_TYPE_TEXT,
-                                 &atk_text_info);
-    gail_cell_type_add_action_interface (type);
-  }
+      type = g_type_register_static (GAIL_TYPE_RENDERER_CELL,
+                                     "GailTextCell", &tinfo, 0);
+      g_type_add_interface_static (type, ATK_TYPE_TEXT,
+                                   &atk_text_info);
+      gail_cell_type_add_action_interface (type);
+    }
   return type;
 }
 
@@ -157,12 +168,15 @@ static void
 gail_text_cell_class_init (GailTextCellClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
-  GailRendererCellClass *renderer_cell_class = GAIL_RENDERER_CELL_CLASS(klass);
+  AtkObjectClass *atk_object_class = ATK_OBJECT_CLASS (klass);
+  GailRendererCellClass *renderer_cell_class = GAIL_RENDERER_CELL_CLASS (klass);
 
-  parent_class = g_type_class_ref (GAIL_TYPE_RENDERER_CELL);
+  parent_class = g_type_class_peek_parent (klass);
   renderer_cell_class->update_cache = gail_text_cell_update_cache;
   renderer_cell_class->property_list = gail_text_cell_property_list;
 
+  atk_object_class->get_name = gail_text_cell_get_name;
+
   gobject_class->finalize = gail_text_cell_finalize;
 }
 
@@ -174,9 +188,9 @@ gail_text_cell_init (GailTextCell *text_cell)
   text_cell->cell_text = NULL;
   text_cell->caret_pos = 0;
   text_cell->cell_length = 0;
-  text_cell->texthelper = gail_text_helper_new();
-  atk_state_set_add_state (GAIL_CELL(text_cell)->state_set,
-    ATK_STATE_SINGLE_LINE);
+  text_cell->textutil = gail_text_util_new ();
+  atk_state_set_add_state (GAIL_CELL (text_cell)->state_set,
+                           ATK_STATE_SINGLE_LINE);
 }
 
 AtkObject* 
@@ -196,6 +210,8 @@ gail_text_cell_new (void)
   cell = GAIL_RENDERER_CELL(object);
 
   cell->renderer = gtk_cell_renderer_text_new ();
+  g_object_ref (cell->renderer);
+  gtk_object_sink (GTK_OBJECT (cell->renderer));
   return atk_object;
 }
 
@@ -204,69 +220,91 @@ gail_text_cell_finalize (GObject            *object)
 {
   GailTextCell *text_cell = GAIL_TEXT_CELL (object);
 
-  g_object_unref (text_cell->texthelper);
+  g_object_unref (text_cell->textutil);
   g_free (text_cell->cell_text);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static G_CONST_RETURN gchar*
+gail_text_cell_get_name (AtkObject *atk_obj)
+{
+  if (atk_obj->name)
+    return atk_obj->name;
+  else
+    {
+      GailTextCell *text_cell = GAIL_TEXT_CELL (atk_obj);
+
+      return text_cell->cell_text;
+    }
+}
+
 static gboolean
 gail_text_cell_update_cache (GailRendererCell *cell,
-  gboolean emit_change_signal)
+                             gboolean         emit_change_signal)
 {
-  GailTextCell *text_cell = GAIL_TEXT_CELL(cell);
+  GailTextCell *text_cell = GAIL_TEXT_CELL (cell);
+  AtkObject *obj = ATK_OBJECT (cell);
   gboolean rv = FALSE;
   gint temp_length;
   gchar *new_cache;
 
-  g_object_get (G_OBJECT(cell->renderer), "text", &new_cache, NULL);
+  g_object_get (G_OBJECT (cell->renderer), "text", &new_cache, NULL);
 
   if (text_cell->cell_text)
-  {
-   /*
-    * If the new value is NULL and the old value isn't NULL, then the
-    * value has changed.
-    */
-    if (new_cache == NULL ||
-        g_strcasecmp (text_cell->cell_text, new_cache))
     {
-      g_free (text_cell->cell_text);
-      temp_length = text_cell->cell_length;
-      text_cell->cell_text = NULL;
-      text_cell->cell_length = 0;
-      g_signal_emit_by_name (G_OBJECT(cell), "text-changed::delete",
-			     0, temp_length);
+     /*
+      * If the new value is NULL and the old value isn't NULL, then the
+      * value has changed.
+      */
+      if (new_cache == NULL ||
+          g_strcasecmp (text_cell->cell_text, new_cache))
+        {
+          g_free (text_cell->cell_text);
+          temp_length = text_cell->cell_length;
+          text_cell->cell_text = NULL;
+          text_cell->cell_length = 0;
+          if (emit_change_signal)
+            {
+              g_signal_emit_by_name (cell, "text_changed::delete", 0, temp_length);
+              if (obj->name == NULL)
+                g_object_notify (G_OBJECT (obj), "accessible-name");
+            }
+          if (new_cache)
+            rv = TRUE;
+        }
     }
-  }
   else
-  {
     rv = TRUE;
-  }
 
   if (rv)
-  {
-    if (new_cache == NULL)
-      {
-	text_cell->cell_text = NULL;
-	text_cell->cell_length = 0;
-      }
-    else
-      {
-	text_cell->cell_text = g_strdup (new_cache);
-	text_cell->cell_length = g_utf8_strlen (new_cache, -1);
-      }
-  }
+    {
+      if (new_cache == NULL)
+        {
+          text_cell->cell_text = g_strdup ("");
+          text_cell->cell_length = 0;
+        }
+      else
+        {
+          text_cell->cell_text = g_strdup (new_cache);
+          text_cell->cell_length = g_utf8_strlen (new_cache, -1);
+        }
+    }
 
-  gail_text_helper_text_setup(GAIL_TEXT_CELL(cell)->texthelper,
-    GAIL_TEXT_CELL(cell)->cell_text);
+  g_free (new_cache);
+  gail_text_util_text_setup (text_cell->textutil, text_cell->cell_text);
   
   if (rv)
-    g_signal_emit_by_name (G_OBJECT (cell), "text-changed::insert",
-			   0, text_cell->cell_length);
+    {
+      if (emit_change_signal)
+        {
+          g_signal_emit_by_name (cell, "text_changed::insert",
+  			         0, text_cell->cell_length);
 
-  gail_text_helper_text_setup(GAIL_TEXT_CELL(cell)->texthelper,
-    GAIL_TEXT_CELL(cell)->cell_text);
-  
+          if (obj->name == NULL)
+            g_object_notify (G_OBJECT (obj), "accessible-name");
+        }
+    }
   return rv;
 }
 
@@ -282,7 +320,8 @@ atk_text_interface_init (AtkTextIface *iface)
   iface->get_character_count = gail_text_cell_get_character_count;
   iface->get_caret_offset = gail_text_cell_get_caret_offset;
   iface->set_caret_offset = gail_text_cell_set_caret_offset;
-  iface->ref_run_attributes = gail_text_cell_ref_run_attributes;
+  iface->get_run_attributes = gail_text_cell_get_run_attributes;
+  iface->get_default_attributes = gail_text_cell_get_default_attributes;
   iface->get_character_extents = gail_text_cell_get_character_extents;
   iface->get_offset_at_point = gail_text_cell_get_offset_at_point;
 }
@@ -292,15 +331,11 @@ gail_text_cell_get_text (AtkText *text,
                          gint    start_pos,
                          gint    end_pos)
 {
-  if (GAIL_TEXT_CELL(text)->cell_text)
-  {
-    return gail_text_helper_get_substring (GAIL_TEXT_CELL(text)->texthelper,
-            &start_pos, &end_pos);
-  }
+  if (GAIL_TEXT_CELL (text)->cell_text)
+    return gail_text_util_get_substring (GAIL_TEXT_CELL (text)->textutil,
+              start_pos, end_pos);
   else
-  {
     return g_strdup ("");
-  }
 }
 
 static gchar* 
@@ -310,17 +345,9 @@ gail_text_cell_get_text_before_offset (AtkText         *text,
                                        gint            *start_offset,
                                        gint            *end_offset)
 {
-  if (boundary_type == ATK_TEXT_BOUNDARY_CHAR)
-  {
-    return gail_text_helper_boundary_char (GAIL_TEXT_CELL(text)->texthelper,
-                                           offset, start_offset, end_offset);
-  }
-  else
-  {
-    return gail_text_helper_switch_boundary_type (GAIL_TEXT_CELL(text)->texthelper,
+  return gail_text_util_get_text (GAIL_TEXT_CELL (text)->textutil,
         NULL, GAIL_BEFORE_OFFSET, boundary_type, offset, start_offset,
         end_offset);
-  }
 }
 
 static gchar* 
@@ -330,16 +357,8 @@ gail_text_cell_get_text_at_offset (AtkText         *text,
                                    gint            *start_offset,
                                    gint            *end_offset)
 {
-  if (boundary_type == ATK_TEXT_BOUNDARY_CHAR)
-  {
-    return gail_text_helper_boundary_char (GAIL_TEXT_CELL(text)->texthelper,
-                                           offset, start_offset, end_offset);
-  }
-  else
-  {
-    return gail_text_helper_switch_boundary_type (GAIL_TEXT_CELL(text)->texthelper,
+  return gail_text_util_get_text (GAIL_TEXT_CELL (text)->textutil,
         NULL, GAIL_AT_OFFSET, boundary_type, offset, start_offset, end_offset);
-  }
 }
 
 static gchar* 
@@ -349,183 +368,110 @@ gail_text_cell_get_text_after_offset (AtkText         *text,
                                       gint            *start_offset,
                                       gint            *end_offset)
 {
-  if (boundary_type == ATK_TEXT_BOUNDARY_CHAR)
-  {
-    return gail_text_helper_boundary_char (GAIL_TEXT_CELL(text)->texthelper,
-        offset, start_offset, end_offset);
-  }
-  else
-  {
-    return gail_text_helper_switch_boundary_type (GAIL_TEXT_CELL(text)->texthelper,
+  return gail_text_util_get_text (GAIL_TEXT_CELL (text)->textutil,
         NULL, GAIL_AFTER_OFFSET, boundary_type, offset, start_offset,
         end_offset);
-  }
 }
 
-static gint gail_text_cell_get_character_count (AtkText *text)
+static gint 
+gail_text_cell_get_character_count (AtkText *text)
 {
-  if (GAIL_TEXT_CELL(text)->cell_text != NULL)
-    return GAIL_TEXT_CELL(text)->cell_length;
+  if (GAIL_TEXT_CELL (text)->cell_text != NULL)
+    return GAIL_TEXT_CELL (text)->cell_length;
   else
     return 0;
 }
 
-static gint gail_text_cell_get_caret_offset (AtkText *text)
+static gint 
+gail_text_cell_get_caret_offset (AtkText *text)
 {
-  return GAIL_TEXT_CELL(text)->caret_pos;
+  return GAIL_TEXT_CELL (text)->caret_pos;
 }
 
-static gboolean gail_text_cell_set_caret_offset (AtkText *text, gint offset)
+static gboolean 
+gail_text_cell_set_caret_offset (AtkText *text,
+                                 gint    offset)
 {
-  if (GAIL_TEXT_CELL(text)->cell_text == NULL)
-  {
+  GailTextCell *text_cell = GAIL_TEXT_CELL (text);
+
+  if (text_cell->cell_text == NULL)
     return FALSE;
-  }
   else
-  {
-
-    /* Only set the caret within the bounds and if it is to a new position. */
-    if (offset >= 0 && offset <= GAIL_TEXT_CELL(text)->cell_length &&
-        offset != GAIL_TEXT_CELL(text)->caret_pos)
     {
-      GAIL_TEXT_CELL(text)->caret_pos = offset;
 
-      /* emit the signal */
-      g_signal_emit_by_name(G_OBJECT(text), "text-caret-moved", offset);
-      return TRUE;
+      /* Only set the caret within the bounds and if it is to a new position. */
+      if (offset >= 0 && 
+          offset <= text_cell->cell_length &&
+          offset != text_cell->caret_pos)
+        {
+          text_cell->caret_pos = offset;
+
+          /* emit the signal */
+          g_signal_emit_by_name (text, "text_caret_moved", offset);
+          return TRUE;
+        }
+      else
+        return FALSE;
     }
-    else
-    {
-      return FALSE;
-    }
-  }
 }
 
 static AtkAttributeSet*
-gail_text_cell_ref_run_attributes (AtkText *text,
-                                  gint    offset,
-                                  gint    *start_offset,
-                                  gint    *end_offset) 
+gail_text_cell_get_run_attributes (AtkText *text,
+                                  gint     offset,
+                                  gint     *start_offset,
+                                  gint     *end_offset) 
 {
   GailRendererCell *gail_renderer; 
   GtkCellRendererText *gtk_renderer;
   AtkAttributeSet *attrib_set = NULL;
-  PangoAttrIterator *iter;
-  gchar *value = NULL;
-  PangoAttrLanguage *pango_lang;
-  PangoFontMask mask;
-  static const gchar *style[] = {"normal",
-                                 "oblique",
-                                 "italic"};
-  static const gchar *variant[] = {"normal",
-                                   "small_caps"};
-  static const gchar *stretch[] = {"ultra_condensed",
-                                   "extra_condensed",
-				   "condensed",
-                                   "semi_condensed",
-                                   "normal",
-                                   "semi_expanded",
-                                   "expanded",
-                                   "extra_expanded",
-                                   "ultra_expanded"};
+  PangoLayout *layout;
+  AtkObject *parent;
+  GtkWidget *widget;
 
-  gail_renderer = GAIL_RENDERER_CELL(text);
-  gtk_renderer = GTK_CELL_RENDERER_TEXT(gail_renderer->renderer);
+  gail_renderer = GAIL_RENDERER_CELL (text);
+  gtk_renderer = GTK_CELL_RENDERER_TEXT (gail_renderer->renderer);
 
-  *start_offset = 0;
-  *end_offset = gail_text_cell_get_character_count(text) - 1;
-
-  mask = pango_font_description_get_set_fields (gtk_renderer->font);
-
-  if (mask & PANGO_FONT_MASK_FAMILY)
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_FAMILY_NAME,
-      pango_font_description_get_family (gtk_renderer->font));
-    
-  if (mask & PANGO_FONT_MASK_STYLE)
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_STYLE,
-      style[pango_font_description_get_style (gtk_renderer->font)]);
-
-  if (mask & PANGO_FONT_MASK_VARIANT)
-  gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_VARIANT,
-      variant[pango_font_description_get_variant (gtk_renderer->font)]);
-
-  if (mask & PANGO_FONT_MASK_WEIGHT)
-  {
-    value = g_strdup_printf("%i", pango_font_description_get_weight (gtk_renderer->font));
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_WEIGHT, value);
-  }
+  parent = atk_object_get_parent (ATK_OBJECT (text));
+  if (GAIL_IS_CONTAINER_CELL (parent))
+    parent = atk_object_get_parent (parent);
+  g_return_val_if_fail (GAIL_IS_CELL_PARENT (parent), NULL);
+  widget = GTK_ACCESSIBLE (parent)->widget;
+  layout = create_pango_layout (gtk_renderer, widget),
+  attrib_set = gail_misc_layout_get_run_attributes (attrib_set, 
+                                                    layout,
+                                                    gtk_renderer->text,
+                                                    offset,
+                                                    start_offset,
+                                                    end_offset);
+  g_object_unref (G_OBJECT (layout));
   
-  if (mask & PANGO_FONT_MASK_STRETCH)
-  {
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_STRETCH,
-      stretch[pango_font_description_get_stretch (gtk_renderer->font)]);
-  }
-  
-  if (mask & PANGO_FONT_MASK_SIZE)
-  {
-    value = g_strdup_printf("%i", pango_font_description_get_size (gtk_renderer->font));
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_SIZE, value);
-  }
+  return attrib_set;
+}
 
-  if (gtk_renderer->scale_set)
-  {
-    value = g_strdup_printf("%g", gtk_renderer->font_scale);
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_SCALE, value);
-  }
+static AtkAttributeSet*
+gail_text_cell_get_default_attributes (AtkText	*text)
+{
+  GailRendererCell *gail_renderer; 
+  GtkCellRendererText *gtk_renderer;
+  AtkAttributeSet *attrib_set = NULL;
+  PangoLayout *layout;
+  AtkObject *parent;
+  GtkWidget *widget;
 
-  if (gtk_renderer->foreground_set)
-  {
-    value = g_strdup_printf("%u,%u,%u", gtk_renderer->foreground.red,
-      gtk_renderer->foreground.green, gtk_renderer->foreground.blue);
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_FG_COLOR, value);
-  }
+  gail_renderer = GAIL_RENDERER_CELL (text);
+  gtk_renderer = GTK_CELL_RENDERER_TEXT (gail_renderer->renderer);
 
-  if (gtk_renderer->background_set)
-  {
-    value = g_strdup_printf("%u,%u,%u", gtk_renderer->background.red,
-      gtk_renderer->background.green, gtk_renderer->background.blue);
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_BG_COLOR, value);
-  }
+  parent = atk_object_get_parent (ATK_OBJECT (text));
+  if (GAIL_IS_CONTAINER_CELL (parent))
+    parent = atk_object_get_parent (parent);
+  g_return_val_if_fail (GAIL_IS_CELL_PARENT (parent), NULL);
+  widget = GTK_ACCESSIBLE (parent)->widget;
+  layout = create_pango_layout (gtk_renderer, widget),
 
-  if (gtk_renderer->underline_set)
-  {
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_UNDERLINE,
-    "true");
-  }
-
-  if (gtk_renderer->rise_set)
-  {
-    value = g_strdup_printf("%i", gtk_renderer->rise);
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_RISE, value);
-  }
-
-  if (gtk_renderer->strikethrough_set)
-  {
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_STRIKETHROUGH,
-    "true");
-  }
-  
-  if (gtk_renderer->editable_set)
-  {
-    gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_EDITABLE,
-    "true");
-  }
-
-  if (gtk_renderer->extra_attrs != NULL)
-  {
-    iter = pango_attr_list_get_iterator(gtk_renderer->extra_attrs);
-    
-    if ((pango_lang = (PangoAttrLanguage*) pango_attr_iterator_get(iter,
-      PANGO_ATTR_LANGUAGE)) != NULL)
-    {
-      gail_text_helper_add_attribute(attrib_set, ATK_ATTRIBUTE_LANGUAGE,
-        pango_language_to_string(pango_lang->value));
-    }
-  }
-
-  if (value)
-    g_free(value);
-  
+  attrib_set = gail_misc_get_default_attributes (attrib_set, 
+                                                 layout,
+                                                 widget);
+  g_object_unref (G_OBJECT (layout));
   return attrib_set;
 }
 
@@ -544,17 +490,6 @@ create_pango_layout(GtkCellRendererText *gtk_renderer,
   PangoUnderline uline;
   PangoFontMask mask;
 
-  /* An internal convenience function to
-   * add a PangoAttribute to a PangoAttrList
-   */
-  void add_attr (PangoAttrList  *attr_list,
-                 PangoAttribute *attr)
-  {
-    attr->start_index = 0;
-    attr->end_index = G_MAXINT;
-    pango_attr_list_insert (attr_list, attr);
-  }
-  
   layout = gtk_widget_create_pango_layout (widget, gtk_renderer->text);
 
   if (gtk_renderer->extra_attrs)
@@ -563,12 +498,12 @@ create_pango_layout(GtkCellRendererText *gtk_renderer,
     attr_list = pango_attr_list_new ();
 
   if (gtk_renderer->foreground_set)
-  {
-    PangoColor color;
-    color = gtk_renderer->foreground;
-    add_attr (attr_list, pango_attr_foreground_new (color.red,
-      color.green, color.blue));
-  }
+    {
+      PangoColor color;
+      color = gtk_renderer->foreground;
+      add_attr (attr_list, pango_attr_foreground_new (color.red,
+                                                      color.green, color.blue));
+    }
 
   if (gtk_renderer->strikethrough_set)
     add_attr (attr_list,
@@ -605,10 +540,8 @@ create_pango_layout(GtkCellRendererText *gtk_renderer,
     uline = PANGO_UNDERLINE_NONE;
 
   if (uline != PANGO_UNDERLINE_NONE)
-  {
     add_attr (attr_list,
       pango_attr_underline_new (gtk_renderer->underline_style));
-  }
 
   if (gtk_renderer->rise_set)
     add_attr (attr_list, pango_attr_rise_new (gtk_renderer->rise));
@@ -620,6 +553,15 @@ create_pango_layout(GtkCellRendererText *gtk_renderer,
   return layout;
 }
 
+static void 
+add_attr (PangoAttrList  *attr_list,
+         PangoAttribute *attr)
+{
+  attr->start_index = 0;
+  attr->end_index = G_MAXINT;
+  pango_attr_list_insert (attr_list, attr);
+}
+
 static void      
 gail_text_cell_get_character_extents (AtkText          *text,
                                       gint             offset,
@@ -639,21 +581,21 @@ gail_text_cell_get_character_extents (AtkText          *text,
   gint x_offset, y_offset, index, cell_height, cell_width;
 
   if (!GAIL_TEXT_CELL (text)->cell_text)
-  {
-    *x = *y = *height = *width = 0;
-    return;
-  }
+    {
+      *x = *y = *height = *width = 0;
+      return;
+    }
   if (offset < 0 || offset >= GAIL_TEXT_CELL (text)->cell_length)
-  {
-    *x = *y = *height = *width = 0;
-    return;
-  }
+    {
+      *x = *y = *height = *width = 0;
+      return;
+    }
   gail_renderer = GAIL_RENDERER_CELL (text);
   gtk_renderer = GTK_CELL_RENDERER_TEXT (gail_renderer->renderer);
   /*
    * Thus would be inconsistent with the cache
    */
-  g_return_if_fail (gtk_renderer->text);
+  gail_return_if_fail (gtk_renderer->text);
 
   parent = atk_object_get_parent (ATK_OBJECT (text));
   if (GAIL_IS_CONTAINER_CELL (parent))
@@ -671,9 +613,11 @@ gail_text_cell_get_character_extents (AtkText          *text,
     offset) - gtk_renderer->text;
   pango_layout_index_to_pos (layout, index, &char_rect); 
 
-  gail_text_helper_get_extents_from_pango_rectangle (widget,
-      &char_rect, x_offset + rendered_rect.x,
-      y_offset + rendered_rect.y, x, y, width, height, coords);
+  gail_misc_get_extents_from_pango_rectangle (widget,
+      &char_rect,
+      x_offset + rendered_rect.x + gail_renderer->renderer->xpad,
+      y_offset + rendered_rect.y + gail_renderer->renderer->ypad,
+      x, y, width, height, coords);
   g_object_unref (layout);
   return;
 } 
@@ -690,12 +634,11 @@ gail_text_cell_get_offset_at_point (AtkText          *text,
   GtkWidget *widget;
   GdkRectangle rendered_rect;
   PangoLayout *layout;
-  gint x_offset, y_offset, index, cell_height, cell_width;
+  gint x_offset, y_offset, index;
  
   if (!GAIL_TEXT_CELL (text)->cell_text)
-  {
     return -1;
-  } 
+
   gail_renderer = GAIL_RENDERER_CELL (text);
   gtk_renderer = GTK_CELL_RENDERER_TEXT (gail_renderer->renderer);
   parent = atk_object_get_parent (ATK_OBJECT (text));
@@ -709,24 +652,26 @@ gail_text_cell_get_offset_at_point (AtkText          *text,
   g_return_val_if_fail (GAIL_IS_CELL_PARENT (parent), -1);
   gail_cell_parent_get_cell_area (GAIL_CELL_PARENT (parent), GAIL_CELL (text),
                                   &rendered_rect);
-  gtk_cell_renderer_get_size(GTK_CELL_RENDERER (gtk_renderer), widget,
-     &rendered_rect, &x_offset, &y_offset, &cell_width, &cell_height);
+  gtk_cell_renderer_get_size (GTK_CELL_RENDERER (gtk_renderer), widget,
+     &rendered_rect, &x_offset, &y_offset, NULL, NULL);
 
   layout = create_pango_layout (gtk_renderer, widget);
    
-  index = gail_text_helper_get_index_at_point_in_layout (widget, layout,
-        x_offset + rendered_rect.x, y_offset + rendered_rect.y,
+  index = gail_misc_get_index_at_point_in_layout (widget, layout,
+        x_offset + rendered_rect.x + gail_renderer->renderer->xpad,
+        y_offset + rendered_rect.y + gail_renderer->renderer->ypad,
         x, y, coords);
   g_object_unref (layout);
   if (index == -1)
-  {
-    return index;  
-  }
+    {
+      if (coords == ATK_XY_WINDOW || coords == ATK_XY_SCREEN)
+        return g_utf8_strlen (gtk_renderer->text, -1);
+    
+      return index;  
+    }
   else
-  {
     return g_utf8_pointer_to_offset (gtk_renderer->text,
        gtk_renderer->text + index);  
-  }
 }
 
 static gunichar 
@@ -734,13 +679,18 @@ gail_text_cell_get_character_at_offset (AtkText       *text,
                                         gint          offset)
 {
   gchar *index;
+  gchar *string;
 
-  if (GAIL_TEXT_CELL(text)->cell_text == NULL)
-  {
+  string = GAIL_TEXT_CELL(text)->cell_text;
+
+  if (!string)
     return '\0';
-  }
-  index = g_utf8_offset_to_pointer(GAIL_TEXT_CELL(text)->cell_text, offset);
 
-  return g_utf8_get_char(index);
+  if (offset >= g_utf8_strlen (string, -1))
+    return '\0';
+
+  index = g_utf8_offset_to_pointer (string, offset);
+
+  return g_utf8_get_char (index);
 }
 
